@@ -8,8 +8,8 @@ exec perl -w -x $0 ${1+"$@"} # -*- mode: perl; perl-indent-level: 2; -*-
 ###                                                        ###
 ##############################################################
 
-## $Revision: 2.18 $
-## $Date: 2000/08/29 22:50:41 $
+## $Revision: 2.29 $
+## $Date: 2000/09/06 16:57:18 $
 ## $Author: kfogel $
 ##
 ##   (C) 1999 Karl Fogel <kfogel@red-bean.com>, under the GNU GPL.
@@ -78,7 +78,7 @@ use File::Basename;
 my $Log_Source_Command = "cvs log";
 
 # In case we have to print it out:
-my $VERSION = '$Revision: 2.18 $';
+my $VERSION = '$Revision: 2.29 $';
 $VERSION =~ s/\S+\s+(\S+)\s+\S+/$1/;
 
 ## Vars set by options:
@@ -110,7 +110,9 @@ my $Prune_Empty_Msgs = 0;
 # Don't call Text::Wrap on the body of the message
 my $No_Wrap = 0;
 
-# Separates header from log message
+# Separates header from log message.  Code assumes it is either " " or
+# "\n\n", so if there's ever an option to set it to something else,
+# make sure to go through all conditionals that use this var.
 my $After_Header = " ";
 
 # Format more for programs than for humans.
@@ -221,6 +223,11 @@ sub derive_change_log ()
   my %branch_numbers;    # Save some revisions for @Follow_Branches
   my @branch_roots;      # For showing which files are branch ancestors.
 
+  # Bleargh.  Compensate for a deficiency of custom wrapping.
+  if (($After_Header ne " ") and $FSF_Style)
+  {
+    $After_Header .= "\t";
+  }
 
   if (! $Input_From_Stdin) {
     open (LOG_SOURCE, "$Log_Source_Command |")
@@ -685,7 +692,6 @@ sub derive_change_log ()
         while (my ($msg,$qunklist) = each %$mesghash)
         {
           my $files               = &pretty_file_list ($qunklist);
-          my $logtext             = &pretty_msg_text ($msg);
           my $header_line;          # date and author
           my $body;                 # see below
           my $wholething;           # $header_line + $body
@@ -722,28 +728,43 @@ sub derive_change_log ()
           }
 
           # Reshape the body according to user preferences.
-          if ($XML_Output) {
-            $body = $files . $logtext;
+          if ($XML_Output) 
+          {
+            $msg = &preprocess_msg_text ($msg);
+            $body = $files . $msg;
           }
-          elsif ($No_Wrap) {
+          elsif ($No_Wrap) 
+          {
+            $msg = &preprocess_msg_text ($msg);
             $files = wrap ("\t", "	", "$files");
-            $logtext =~ s/\n(.*)/\n\t$1/g;
+            $msg =~ s/\n(.*)/\n\t$1/g;
             unless ($After_Header eq " ") {
-              $logtext =~ s/^(.*)/\t$1/g;
+              $msg =~ s/^(.*)/\t$1/g;
             }
-            $body = $files . $After_Header . $logtext;
+            $body = $files . $After_Header . $msg;
           }
-          else {
-            $body = $files . $After_Header . $logtext;
-            $body = wrap ("\t", "        ", "$body");
-
+          else  # do wrapping, either FSF-style or regular
+          {
             if ($FSF_Style)
             {
-              # Compensate for earlier tweaks made by pretty_msg_text()
-              # to protect FSF-style ChangeLog formatting.  See code
-              # there for details.
-              $body =~ s/\n\t(\n\t\([a-zA-Z0-9,_ \.\t\n-]+\): )/$1/g;
-              $body =~ s/\n\t(\n\t\*[ \t]+[a-zA-Z0-9,_ \.\t\n\?\!-]+)/\n$1/g;
+              $files = wrap ("\t", "        ", "$files");
+              
+              my $files_last_line_len = 0;
+              if ($After_Header eq " ")
+              {
+                $files_last_line_len = &last_line_len ($files);
+                $files_last_line_len += 1;  # for $After_Header
+              }
+              
+              $msg = &wrap_log_entry
+                  ($msg, "\t", 69 - $files_last_line_len, 69);
+              $body = $files . $After_Header . $msg;
+            }
+            else  # not FSF-style
+            {
+              $msg = &preprocess_msg_text ($msg);
+              $body = $files . $After_Header . $msg;
+              $body = wrap ("\t", "        ", "$body");
             }
           }
 
@@ -1118,7 +1139,7 @@ sub common_path_prefix ()
 }
 
 
-sub pretty_msg_text ()
+sub preprocess_msg_text ()
 {
   my $text = shift;
 
@@ -1135,37 +1156,6 @@ sub pretty_msg_text ()
   }
   elsif (! $No_Wrap)
   {
-    if ($FSF_Style)
-    {
-      # Tweak to make FSF-style ChangeLog syntax work better.  There are
-      # two special things in FSF-style ChangeLog format: the
-      # asterisked-file-list thing, and the paren-grouped-symbol
-      # thing.  Here's an example showing both: 
-      #
-      #    * configure.in: Include lib/Makefile in the list of
-      #    Makefiles to produce.
-      #    * include/cerror.h (cereal_errno_t): Add new error
-      #    code, CEREAL_ERR_FS_GENERAL.
-      #    * libcereal.c, libmilk.c (fill_bowl): fix overflow
-      #    problem by passing and checking new rim_height
-      #    argument.
-      #    * libcondiments.c (raisin_fill, coat_with_sugar,
-      #    make_advertisement): don't use real raisins, use
-      #    artificial ones (the users won't know anyway).
-      #    (make_nuts, display_nuts): use fake nuts, too.
-      # 
-      # If we just wrap() that entry, it will get garbled.  So below
-      # we add some extra newline shims before the relevent asterisks
-      # and open parens, to prevent wrapping, and the shims get
-      # removed after the wrap() call in derive_change_log().
-      
-      # Protect paren-grouped-symbols:
-      $text =~ s/(\([a-zA-Z0-9,_ \.\t\n\?\!-]+\): )/\n$1/g;
-
-      # Protect asterisked-file-lists:
-      $text =~ s/(\*[ \t]+[a-zA-Z0-9,_ \.\t\n\?\!-]+)/\n$1/g;
-    }
-
     # Strip off lone newlines, but only for lines that don't begin with
     # whitespace or a mail-quoting character, since we want to preserve
     # that kind of formatting.  Also don't strip newlines that follow a
@@ -1182,6 +1172,159 @@ sub pretty_msg_text ()
 }
 
 
+sub last_line_len ()
+{
+  my $files_list = shift;
+  my @lines = split (/\n/, $files_list);
+  my $last_line = pop (@lines);
+  return length ($last_line);
+}
+
+
+# A custom wrap function, sensitive to some common constructs used in
+# log entries.
+sub wrap_log_entry ()
+{
+  my $text = shift;                # The text to wrap.
+  my $left_pad_str = shift;        # String to pad with on the left.
+
+  # These do NOT take left_pad_str into account:
+  my $length_remaining = shift;    # Amount left on current line.
+  my $max_line_length  = shift;    # Amount left for a blank line.
+
+  my $wrapped_text = "";           # The accumulating wrapped entry.
+  my $indentation = "";            # Inherited indentation from prev line.
+
+  my $first_time = 1;              # Is this the first iteration of the loop?
+
+  my @lines = split (/\n/, $text);
+  while (@lines)   # Don't use `foreach' here, it won't work.
+  {
+    my $this_line = shift (@lines);
+    chomp $this_line;
+
+    # First, if it matches any of the line-start regexps, then print a
+    # newline now...
+    if (($this_line =~ /^(\s*)\*\s+[a-zA-Z0-9]/)
+        || ($this_line =~ /^(\s*)\* [a-zA-Z0-9_\.\/\+-]+/)
+        || ($this_line =~ /^(\s*)\([a-zA-Z0-9_\.\/\+-]+(\)|,\s*)/)
+        || ($this_line =~ /^(\s*)- +/)
+        || ($this_line =~ /^(\s*)[^\s]+:\s*$/)
+        || ($this_line =~ /^(\s*)\*\) +/)
+        || ($this_line =~ /^(\s*)[a-zA-Z0-9](\)|\.|\:) +/))
+    {
+      # Make a line break immediately, unless header separator is set
+      # and this line is the first line in the entry, in which case
+      # we're getting the blank line for free already and shouldn't
+      # add an extra one.
+      unless (($After_Header ne " ") and ($first_time)) {
+        $wrapped_text .= "\n${left_pad_str}";
+      }
+      
+      $indentation = $1;
+      $length_remaining = $max_line_length - (length ($indentation));
+    }
+
+    # Now that any indentation has been preserved, strip off leading
+    # whitespace, so up-folding has no ugly side-effects.
+    $this_line =~ s/^\s*//;
+
+    # Accumulate the line, and adjust parameters for next line.
+    my $this_len = length ($this_line);
+    if ($this_len == 0)
+    {
+      # Blank lines should be preserved, and should cancel any
+      # indentation level.
+      $this_line = "\n${left_pad_str}";
+      $indentation = "";
+      $length_remaining = $max_line_length;
+    }
+    elsif ($this_len >= $length_remaining)
+    {
+      # Walk backwards from the end.  At first acceptable spot, break
+      # a new line.
+      my $idx = $length_remaining - 1;
+      if ($idx < 0) { $idx = 0 };
+      while ($idx > 0)
+      {
+        if (substr ($this_line, $idx, 1) =~ /\s/)
+        {
+          my $line_now = substr ($this_line, 0, $idx);
+          my $next_line = substr ($this_line, $idx);
+          $this_line = $line_now;
+          
+          # Clean whitespace off the end.
+          chomp $this_line;
+
+          # The current line is ready to be printed.
+          $this_line .= "\n${left_pad_str}";
+
+          # Make sure the next line is allowed full room.
+          $length_remaining = $max_line_length - (length ($indentation));
+
+          # Strip next_line, but then preserve any indentation.
+          $next_line =~ s/^\s*//;
+          $next_line = $indentation . $next_line;
+
+          unshift (@lines, $next_line);
+          last;
+        }
+        else
+        {
+          $idx--;
+        }
+      }
+
+      # If we just bottomed out, the line is too long, so just break
+      # it where the original text does.
+      if ($idx == 0)
+      {
+        if ($length_remaining == ($max_line_length - (length ($indentation))))
+        {
+          # The line is simply too long -- there is no hope of ever
+          # breaking it nicely, so just insert it verbatim, with
+          # appropriate padding.
+          $this_line = "\n${left_pad_str}${this_line}";
+        }
+        else
+        {
+          # Can't break it here, but may be able to on the next round.
+          unshift (@lines, $this_line);
+          $length_remaining = $max_line_length - (length ($indentation));
+          $this_line = "\n${left_pad_str}";
+        }
+      }
+    }
+    else  # $this_len < $length_remaining, so tack on what we can.
+    {
+      # Leave a note for the next iteration.
+      $length_remaining = $length_remaining - $this_len;
+
+      if ($this_line =~ /\.$/)
+      {
+        $this_line .= "  ";
+        $length_remaining -= 2;
+      }
+      else  # not a sentence end
+      {
+        $this_line .= " ";
+        $length_remaining -= 1;
+      }
+    }
+
+    # Unconditionally indicate that loop has run at least once.
+    $first_time = 0;
+
+    $wrapped_text .= "${this_line}";
+  }
+
+  # One last bit of padding.
+  $wrapped_text .= "\n";
+
+  return $wrapped_text;
+}
+
+
 sub xml_escape ()
 {
   my $txt = shift;
